# Copyright (C) 2025 Robotec.AI
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pydantic import BaseModel, ConfigDict


class RaiBaseModel(BaseModel):
    model_config = ConfigDict(extra="forbid")


class ROS2BaseModel(RaiBaseModel):
    """
    Base model for messages replicated from ros2

    Attributes:
    _prefix: str
        Prefix of the ros2 message, for example std_msgs/msg
    """

    _prefix: str

    def get_msg_name(self) -> str:
        return f"{self._prefix}/{self.__class__.__name__}"
